/*
    #FCDS - FERCSA's charger and discharger station demo software (Revision A).
    
    -Initial release for my discharger module. If schematic or PCB design not 
    included, please contact me.
    
    Copyright (C) 2017  FERCSA ( fercsa@freemail.hu , www.fercsa.com )

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.





    Welcome to FCDS Station
  
    Hardware coneections.
    Between the master board(Bluepill) and the discharger module:
    GND  - GND
    PA0  - Voltage
    PA1  - Current
    PA2  - PWM
    PA3  - Temp
    3V3  - 3V3

    Between the discharger module and the TP4056 module:
    EN   - Lift off the 8th pin of the TP4056 IC then solder to it.

    Between the discharger module and the thermistor:
    TH1 - Two jumpwire until the 2x8 pin connector or a direct connect to the 
    NTC thermistor.
  
    Sample output from the console:
    #FCDS - FERCSA's charger and discharger station. More info: www.fercsa.com
    #Press 'd' for discharge, 's' for stop.
    PWM: 30000 | Voltage: 3.73V | Current: 0.0A | Capacity: 0.00mAh | Temp: 25C
    PWM: 53878 | Voltage: 3.60V | Current: 0.8A | Capacity: 0.15mAh | Temp: 26C
    PWM: 55779 | Voltage: 3.56V | Current: 1.0A | Capacity: 0.41mAh | Temp: 26C
    PWM: 55687 | Voltage: 3.55V | Current: 1.0A | Capacity: 0.69mAh | Temp: 26C
    PWM: 55613 | Voltage: 3.54V | Current: 1.0A | Capacity: 0.97mAh | Temp: 26C
    PWM: 55524 | Voltage: 3.53V | Current: 1.0A | Capacity: 1.25mAh | Temp: 26C
    PWM: 55461 | Voltage: 3.54V | Current: 1.0A | Capacity: 1.52mAh | Temp: 26C
    PWM: 55417 | Voltage: 3.53V | Current: 1.0A | Capacity: 1.80mAh | Temp: 26C

    Throw me a beer or two if you like it. Cheers!
*/

#define DEBUG                           0
#define VCC                             3.3
#define CELL_VOLTAGE_CUT                3 /* Min. cell voltage */
#define DISCHARGE_CURRENT               1 /* Amper */
#define R_SENSE                         1 /* it should be a 1% resistor with low ppm! */
#define DIVIDER_CELL                    0.0011/* VCC / hand measured voltage on the voltage divider output */
#define DIVIDER_R_SENSE                 0.0010985353/* VCC / hand measured voltage on the voltage divider output */
#define PWM_VIRTUAL_ZERO                30000
#define PWM_HEATUP                      50000

#define PIN_VOLTAGE                     PA0
#define PIN_CURRENT                     PA1
#define PIN_PWM                         PA2
#define PIN_TEMP                        PA3

char command;
float cell_voltage                    = 0;
float r_sense_voltage                 = 0;
float cell_current                    = 0;
float cell_capacity                   = 0;
float temp                            = 0;
int pwm                               = PWM_VIRTUAL_ZERO;
unsigned long serial_prev_millis      = 0;
const long serial_refresh_interval    = 1000;
unsigned long capacity_elapsed_millis = 0;
unsigned long capacity_prev_millis    = 0;


void setup() {
  /* SERIAL */
  Serial.begin(115200);
  /* PWM */
  pinMode(PIN_PWM, PWM);
  pwmWrite(PIN_PWM, pwm);
  /* WE ARE READY TO GO! */
  Serial.println("#FCDS demo ready. Rev.A");
}

void loop(){
  if (Serial.available() > 0) {
    Serial.println();
    Serial.println("#FCDS - FERCSA's charger and discharger station. More info: www.fercsa.com");
    Serial.println("#Press 'd' for discharge, 's' for stop.");
    command = Serial.read();
  }
  
  //STOP AND RESET
  if(command == 's'){
    command = '0';//unset aka do nothing..
    cell_capacity = 0;//reset capacity
    pwm = PWM_VIRTUAL_ZERO;
    pwmWrite(PIN_PWM, pwm);
    Serial.println("#STOP.");
  }

  //DISCHARGE
  if(command == 'd'){
    cell_voltage = get_cell_voltage(PIN_CURRENT, DIVIDER_CELL);
    r_sense_voltage = get_cell_voltage(PIN_VOLTAGE, DIVIDER_R_SENSE);
    cell_current = (cell_voltage - r_sense_voltage) / R_SENSE;
    cell_capacity = get_capacity_sum(cell_current, cell_capacity);
    temp = get_temp();
    if(cell_voltage < CELL_VOLTAGE_CUT){
      Serial.print("#Cell is under ");
      Serial.print(CELL_VOLTAGE_CUT);
      Serial.print("V (");
      Serial.print(cell_voltage);
      Serial.println("V). We should stop.");
      command = 's';
    }
    else{
      display_discharge(cell_voltage, r_sense_voltage, cell_current, cell_capacity, temp);
      adjust_pwm(cell_current, DISCHARGE_CURRENT);
    }
  }
}

/* CELL VOLTAGE */
float get_cell_voltage(byte channel_nr, float divider){
  int cell_analog = 0;
  for(byte i=0;i<10;i++){
    cell_analog += analogRead(channel_nr);
  }
  int cell_analog_average = cell_analog / 10;
  float cell_voltage = cell_analog_average * divider;

  return cell_voltage;
}
/* CELL CAPACITY */
float get_capacity_sum(float cell_current, float cell_capacity){
  capacity_elapsed_millis = millis() - capacity_prev_millis;
  float calculated_cell_capacity = cell_capacity + (cell_current * capacity_elapsed_millis / 3600);
  capacity_prev_millis = millis();
  return calculated_cell_capacity;
}
/* ADJUST PWM */
void adjust_pwm(float cell_current, float target_current){
  if(pwm == PWM_VIRTUAL_ZERO){
    pwm = PWM_HEATUP;//reset to default
  }
  if(cell_current < target_current && pwm < 65535){//for 8bit=255, for 16bit=65535
    pwm += 1;
  }
  else if(cell_current > target_current && pwm > 0){
    pwm -= 1;
  }
  pwmWrite(PIN_PWM, pwm);
}
/* TEMP */
float get_temp(){
  int temp_analog = analogRead(PIN_TEMP);

  long r;
  float t;

  r = (10000 * (4095.0f / float(temp_analog) - 1));

  t = log(r);
  t = 1 / (0.001129148 + (0.000234125 * t) + (0.0000000876741 * t * t * t));
  t = t - 273.15;
  return t;
}
/* DISPLAY DATA */
void display_discharge(float cell_voltage, float r_sense_voltage, float cell_current, float cell_capacity, float temp){

  unsigned long serial_current_millis = millis();

  if(serial_current_millis - serial_prev_millis >= serial_refresh_interval){
    serial_prev_millis = serial_current_millis;
    Serial.print("PWM: ");
    Serial.print(pwm);
    Serial.print(" | Voltage: ");
    Serial.print(cell_voltage);
    Serial.print("V | Current: ");
    Serial.print(cell_current, 1);
    Serial.print("A | Capacity: ");
    Serial.print(cell_capacity);
    Serial.print("mAh | Temp: ");
    Serial.print(temp, 0);
    Serial.print("C");
    Serial.println();
  }
}

